import json
import nbformat as nbf
import os
import pytest
import shutil
import tempfile

from glob import glob
from unittest import mock

from otter.check.logs import EventType, Log, LogEntry
from otter.execute import grade_notebook

from ..utils import TestFileManager, write_ok_test


FILE_MANAGER = TestFileManager(__file__)


@pytest.fixture
def temp_dir():
    d = tempfile.mkdtemp()
    yield d
    shutil.rmtree(d)


def test_kernel_override(temp_dir):
    """
    Tests that ``otter.execute.grade_notebook`` overrides a custom kernel name when indicated.
    """
    nb = nbf.v4.new_notebook(cells=[nbf.v4.new_code_cell("x = 2")])
    nb["metadata"]["kernelspec"] = {"name": "somekernel"}
    subm_path = os.path.join(temp_dir, "submission.ipynb")
    nbf.write(nb, subm_path)

    test_dir = os.path.join(temp_dir, "tests")
    os.makedirs(test_dir)

    write_ok_test(os.path.join(test_dir, "q1.py"), ">>> assert x == 2")

    results = grade_notebook(
        subm_path,
        test_dir=test_dir,
        tests_glob=glob(os.path.join(test_dir, "*.py")),
        ignore_errors=False,
    )

    assert results.notebook["metadata"]["kernelspec"]["name"] == "python3"
    assert results.total == 1


def test_log_execution(temp_dir):
    """
    Test for ``otter.execute.grade_notebook`` when a log is provided.
    """
    log = Log([
        LogEntry(EventType.INIT),
        LogEntry(EventType.CHECK, question="q1"),
        LogEntry(EventType.CHECK, question="q2"),
        LogEntry(EventType.CHECK, question="q3"),
        LogEntry(EventType.BEGIN_EXPORT),
        LogEntry(EventType.END_EXPORT),
    ])

    log.entries[1].shelve({"a": 1, "b": 2})
    log.entries[2].shelve({"a": 1, "b": 2, "c": None})
    log.entries[3].shelve({"a": 2, "b": 2, "c": 3, "d": 4})

    nb = nbf.v4.new_notebook(cells=[
        nbf.v4.new_code_cell("import pandas as pd"),
        nbf.v4.new_code_cell("from pandas import read_csv"),
        nbf.v4.new_code_cell("raise Exception\na, b = 1, 2"),
        nbf.v4.new_code_cell("raise Exception\nc = None"),
        nbf.v4.new_code_cell("raise Exception\na, c, d = 2, 3, 4"),
    ])

    subm_path = os.path.join(temp_dir, "submission.ipynb")
    nbf.write(nb, subm_path)

    test_dir = os.path.join(temp_dir, "tests")
    os.makedirs(test_dir)

    write_ok_test(os.path.join(test_dir, "q1.py"), ">>> assert a == 1 and b == 2")
    write_ok_test(os.path.join(test_dir, "q2.py"), ">>> assert c is None")
    write_ok_test(os.path.join(test_dir, "q3.py"), ">>> assert a == 2 and c == 3 and d == 4")

    results = grade_notebook(subm_path, test_dir=test_dir, log=log, ignore_errors=False)

    with FILE_MANAGER.open("expected_log_results.json") as f:
        expected_results = json.load(f)

    # fix paths to account for the temp directory
    for d in expected_results.values():
        d["path"] = os.path.join(test_dir, os.path.split(d["path"])[1])

    assert results.to_dict() == expected_results


@mock.patch("otter.execute.pickle.load")
def test_results_load_failure_handling(mocked_pickle_load, temp_dir):
    """
    Tests that ``otter.execute.grade_notebook`` handles the case when the results pickle file can't
    be loaded.
    """
    nb = nbf.v4.new_notebook(cells=[nbf.v4.new_code_cell("x = 2")])
    subm_path = os.path.join(temp_dir, "submission.ipynb")
    nbf.write(nb, subm_path)

    test_dir = os.path.join(temp_dir, "tests")
    os.makedirs(test_dir)

    write_ok_test(os.path.join(test_dir, "q1.py"), ">>> assert x == 2")

    mocked_pickle_load.side_effect = Exception("nope")

    results = grade_notebook(
        subm_path,
        test_dir=test_dir,
        tests_glob=glob(os.path.join(test_dir, "*.py")),
        ignore_errors=False,
    )

    assert results.has_catastrophic_failure()



"""Tests for ``otter.grade``"""

import logging
import os
import pandas as pd
import pytest
import re
import shutil
import zipfile

from glob import glob
from python_on_whales import docker
from unittest import mock

from otter.generate import main as generate
from otter.grade import main as grade
from otter.grade.utils import POINTS_POSSIBLE_LABEL
from otter.run.run_autograder.autograder_config import AutograderConfig
from otter.utils import loggers

from ..utils import TestFileManager


ASSIGNMENT_NAME = "otter-grade-test"
FILE_MANAGER = TestFileManager(__file__)
AG_ZIP_PATH = FILE_MANAGER.get_path("autograder.zip")
ZIP_SUBM_PATH = "test/subm.zip"


@pytest.fixture(autouse=True)
def cleanup_output(cleanup_enabled):
    yield
    if cleanup_enabled:
        if os.path.exists("test/final_grades.csv"):
            os.remove("test/final_grades.csv")
        if os.path.exists("test/submission_pdfs"):
            shutil.rmtree("test/submission_pdfs")
        if os.path.exists(ZIP_SUBM_PATH):
            os.remove(ZIP_SUBM_PATH)
        if os.path.exists("test/grading-summaries"):
            shutil.rmtree("test/grading-summaries")


@pytest.fixture(autouse=True, scope="module")
def generate_zip_file_and_patch_build():
    """
    Generate an autograder zip file for use in these tests and patch
    ``otter.grade.containers.build_image`` with ``build_image_with_local_changes``.
    """
    generate(
        tests_dir = FILE_MANAGER.get_path("tests"), 
        requirements = FILE_MANAGER.get_path("requirements.txt"), 
        output_path = AG_ZIP_PATH,
        config = FILE_MANAGER.get_path("otter_config.json"),
        no_environment = True,
    )

    with loggers.level_context(logging.DEBUG):
        yield

    if os.path.isfile(AG_ZIP_PATH):
        os.remove(AG_ZIP_PATH)


@pytest.fixture
def expected_points():
    test_points = {}
    for test_file in glob(FILE_MANAGER.get_path("tests/*.py")):
        env = {}
        with open(test_file) as f:
            exec(f.read(), env)

        test_points[env['test']['name']] = env['test']['points']

    return test_points


@pytest.mark.slow
@pytest.mark.docker
def test_timeout_some_notebooks_finish():
    """
    Check notebook ``1min.ipynb`` is killed due to exceeding the defined timeout while notebook ``10s.ipynb`` is graded;
    The final_grade.csv records everything correctly
    """
    grade_timeout = 30
    grade(
        name = ASSIGNMENT_NAME,
        paths = [FILE_MANAGER.get_path("timeout/")],
        output_dir = "test/",
        autograder = AG_ZIP_PATH,
        containers = 5,
        timeout = grade_timeout,
    )
    df_test = pd.read_csv("test/final_grades.csv")
    assert df_test.iloc[0]["grading_status"] == "--"
    assert df_test.iloc[1]["grading_status"] == "Completed"
    pattern = rf"Executing '[\w.\/-]*test\/test_grade\/files\/timeout\/1min\.ipynb' in docker container timed out in {grade_timeout} seconds"
    assert re.match(pattern, df_test.iloc[2]["grading_status"]) is not None


@pytest.mark.slow
@pytest.mark.docker
def test_timeout_no_notebooks_finish():
    """
    Check notebook ``1min.ipynb`` and ``10s.ipynb are killed due to exceeding the defined timeout;
    The final_grade.csv records everything correctly
    """
    grade_timeout = 5
    grade(
        name = ASSIGNMENT_NAME,
        paths = [FILE_MANAGER.get_path("timeout/")],
        output_dir = "test/",
        autograder = AG_ZIP_PATH,
        containers = 5,
        timeout = grade_timeout,
    )
    df_test = pd.read_csv("test/final_grades.csv")
    pattern1min = rf"Executing '[\w.\/-]*test\/test_grade\/files\/timeout\/1min\.ipynb' in docker container timed out in {grade_timeout} seconds"
    pattern10s = rf"Executing '[\w.\/-]*test\/test_grade\/files\/timeout\/10s\.ipynb' in docker container timed out in {grade_timeout} seconds"
    assert re.match(pattern10s, df_test.iloc[0]["grading_status"]) is not None
    assert re.match(pattern1min, df_test.iloc[1]["grading_status"]) is not None


@pytest.mark.slow
@pytest.mark.docker
def test_network(expected_points):
    """
    Check that the notebook ``network.ipynb`` is unable to do some network requests with disabled
    networking.
    """
    grade(
        name = ASSIGNMENT_NAME,
        paths = [FILE_MANAGER.get_path("network/")],
        output_dir = "test/",
        autograder = AG_ZIP_PATH,
        containers = 5,
        no_network = True,
    )

    df_test = pd.read_csv("test/final_grades.csv")

    # sort by filename
    df_test = df_test.sort_values("file").reset_index(drop=True)

    for _, row in df_test.iterrows():
        for test in expected_points:
            if 'network' == row["file"] and ('q2' in test or 'q3' in test):
                assert row[test] == 0, "{} supposed to fail {} but passed".format(row["file"], test)
            else:
                assert row[test] == expected_points[test], "{} supposed to pass {} but failed".format(row["file"], test)


@pytest.mark.slow
@pytest.mark.docker
def test_notebooks_with_pdfs(expected_points):
    """
    Checks that notebooks are graded correctly and that PDFs are generated.
    """
    grade(
        name = ASSIGNMENT_NAME,
        paths = [FILE_MANAGER.get_path("notebooks/")],
        output_dir = "test/",
        autograder = AG_ZIP_PATH,
        containers = 5,
        pdfs = True,
    )

    # read the output and expected output
    df_test = pd.read_csv("test/final_grades.csv")

    # sort by filename
    df_test = df_test.sort_values("file").reset_index(drop=True)
    df_test["failures"] = df_test["file"].apply(lambda x: [int(n) for n in re.split(r"\D+", x) if len(n) > 0])

    # add score sum cols for tests
    for test in expected_points:
        test_cols = [l for l in df_test.columns if bool(re.search(fr"\b{test}\b", l))]
        df_test[test] = df_test[test_cols].sum(axis=1)

    # check point values
    for _, row in df_test.iterrows():
        for test in expected_points:
            if int(re.sub(r"\D", "", test)) in row["failures"]:
                # q6.py has all_or_nothing set to False, so if the hidden tests fail you should get 2.5 points
                if "6H" in row["file"] and "q6" == test:
                    assert row[test] == 2.5, "{} supposed to fail {} but passed".format(row["file"], test)
                else:
                    assert row[test] == 0, "{} supposed to fail {} but passed".format(row["file"], test)
            else:
                assert row[test] == expected_points[test], "{} supposed to pass {} but failed".format(row["file"], test)

    assert os.path.exists("test/submission_pdfs"), "PDF folder is missing"

    # check that an pdf exists for each submission
    dir1_contents, dir2_contents = (
        [os.path.splitext(f)[0] for f in os.listdir(FILE_MANAGER.get_path("notebooks/")) if not (os.path.isdir(os.path.join(FILE_MANAGER.get_path("notebooks/"), f)))],
        [os.path.splitext(f)[0] for f in os.listdir("test/submission_pdfs") if not (os.path.isdir(os.path.join("test/submission_pdfs", f)))],
    )
    assert sorted(dir1_contents) == sorted(dir2_contents), f"'{FILE_MANAGER.get_path('notebooks/')}' and 'test/submission_pdfs' have different contents"

    # check that the row with point totals for each question exists
    assert any(POINTS_POSSIBLE_LABEL in row for row in df_test.itertuples(index=False))


@mock.patch("otter.grade.launch_containers")
def test_single_notebook_grade(mocked_launch_grade):
    """
    Checks that when a single submission is passed to Otter Grade, it returns the percentage score.
    """
    df = pd.DataFrame([{
        "q1": 2.0,
        "q2": 2.0,
        "q3": 2.0,
        "q4": 1.0,
        "q6": 5.0,
        "q2b": 2.0,
        "q7": 1.0,
        "percent_correct": 1,
        "total_points_earned": 15.0,
        "file": POINTS_POSSIBLE_LABEL,
        "grading_status": "--"
    },{
        "q1": 2.0,
        "q2": 2.0,
        "q3": 2.0,
        "q4": 1.0,
        "q6": 4.0,
        "q2b": 2.0,
        "q7": 1.0,
        "percent_correct": 0.933333,
        "total_points_earned": 14.0,
        "file": "passesAll.ipynb",
        "grading_status": "Completed"
    }])

    notebook_path = FILE_MANAGER.get_path("notebooks/passesAll.ipynb")

    kw_expected = {
        "num_containers": 1,
        "base_image": "ubuntu:22.04",
        "tag": ASSIGNMENT_NAME,
        "no_kill": False,
        "pdf_dir": None,
        "timeout": None,
        "network": True,
        "config": AutograderConfig(),
    }

    mocked_launch_grade.return_value = [df]

    output = grade(
        name = ASSIGNMENT_NAME,
        paths = [notebook_path],
        output_dir = "test/",
        # the value of the autograder argument doesn't matter, it just needs to be a valid file path
        autograder = notebook_path,
        containers = 1,
    )

    mocked_launch_grade.assert_called_with(notebook_path, [notebook_path], **kw_expected)
    assert output == 0.933333


@mock.patch("otter.grade.launch_containers")
def test_config_overrides(mocked_launch_grade):
    """
    Checks that the CLI flags are converted to config overrides correctly.
    """
    mocked_launch_grade.return_value = [pd.DataFrame([{
        "q1": 2.0,
        "q2": 2.0,
        "q3": 2.0,
        "q4": 1.0,
        "q6": 5.0,
        "q2b": 2.0,
        "q7": 1.0,
        "percent_correct": 1,
        "total_points_earned": 15.0,
        "file": POINTS_POSSIBLE_LABEL,
        "grading_status": "--"
    },{
        "q1": 2.0,
        "q2": 2.0,
        "q3": 2.0,
        "q4": 1.0,
        "q6": 5.0,
        "q2b": 2.0,
        "q7": 1.0,
        "percent_correct": 1.0,
        "total_points_earned": 15.0,
        "file": "passesAll.ipynb",
        "grading_status": "Completed"
    }])]

    notebook_path = FILE_MANAGER.get_path("notebooks/passesAll.ipynb")
    grade(
        name = ASSIGNMENT_NAME,
        paths = [notebook_path],
        output_dir = "test/",
        # the value of the autograder argument doesn't matter, it just needs to be a valid file path
        autograder = notebook_path,
        containers = 1,
        pdfs = True,
        ext = "zip",
        debug = True,
    )

    assert mocked_launch_grade.call_args.kwargs["config"].get_user_config() == {
        "zips": True,
        "pdf": True,
        "debug": True,
    }


@pytest.mark.slow
@pytest.mark.docker
def test_config_overrides_integration():
    """
    Checks that overriding otter_config.json configurations with CLI flags works.
    """
    notebook_path = FILE_MANAGER.get_path("notebooks/passesAll.ipynb")
    with zipfile.ZipFile(ZIP_SUBM_PATH, "x") as zf:
        zf.write(notebook_path, arcname="passesAll.ipynb")

    output = grade(
        name = ASSIGNMENT_NAME,
        paths = [ZIP_SUBM_PATH],
        output_dir = "test/",
        # the value of the autograder argument doesn't matter, it just needs to be a valid file path
        autograder = AG_ZIP_PATH,
        ext = "zip",
    )

    assert output == 1.0

    got = pd.read_csv("test/final_grades.csv")
    want = pd.DataFrame([{
        "q1": 0.0,
        "q2": 2.0,
        "q3": 2.0,
        "q4": 1.0,
        "q6": 5.0,
        "q2b": 2.0,
        "q7": 1.0,
        "percent_correct": 1,
        "total_points_earned": 13.0,
        "file": POINTS_POSSIBLE_LABEL,
        "grading_status": "--"
    },{
        "q1": 0.0,
        "q2": 2.0,
        "q3": 2.0,
        "q4": 1.0,
        "q6": 5.0,
        "q2b": 2.0,
        "q7": 1.0,
        "percent_correct": 1.0,
        "total_points_earned": 13.0,
        "file": os.path.splitext(os.path.basename(ZIP_SUBM_PATH))[0],
        "grading_status": "Completed"
    }])

    # Sort the columns by label so the dataframes can be compared with ==.
    got = got.reindex(sorted(got.columns), axis=1)
    want = want.reindex(sorted(want.columns), axis=1)
    assert got.equals(want)


@mock.patch("otter.grade.launch_containers")
def test_grade_summaries(mocked_launch_grade):
    """
    Checks that are grading summaries are written to the disck
    """
    mock_dfs = []
    for filename in os.listdir(FILE_MANAGER.get_path("results")):
        filename = os.path.splitext(filename)[0]
        test_file_path = os.path.join("test/test_grade/files/results", f"{filename}.txt")
        with open(test_file_path, 'r') as test_summary_file:
            mock_dfs.append(pd.DataFrame([{
                "percent_correct": 1.0,
                "total_points_earned": 15.0,
                "file": f"{filename}",
                "summary": test_summary_file.read(),
                "grading_status": "Completed"
            }]))
    mocked_launch_grade.return_value = mock_dfs

    notebook_path = FILE_MANAGER.get_path("notebooks")
    grade(
        name = ASSIGNMENT_NAME,
        paths = [notebook_path],
        output_dir = "test/",
        autograder = AG_ZIP_PATH,
        summaries = True
    )
    for filename in os.listdir(notebook_path):
        filename = os.path.splitext(filename)[0]
        file_path = os.path.join("test/grading-summaries", f"{filename}.txt")
        test_file_path = os.path.join("test/test_grade/files/results", f"{filename}.txt")
        if os.path.isfile(file_path):
            assert os.path.exists(file_path)
            with open(file_path, 'r') as summary_file, open(test_file_path, 'r') as test_summary_file:
                assert summary_file.read() == test_summary_file.read()



"""Tests for ``otter.utils``"""

import pandas as pd

from unittest import mock

from otter.utils import get_variable_type, hide_outputs


@mock.patch("otter.utils.get_ipython")
def test_hide_outputs(mocked_get_ipython):
    """
    Tests for ``otter.utils.hide_outputs``.
    """
    old_formatters = mocked_get_ipython.return_value.display_formatter.formatters
    with hide_outputs():
        assert mocked_get_ipython.return_value.display_formatter.formatters == {}
    assert mocked_get_ipython.return_value.display_formatter.formatters == old_formatters


# Used in test_get_variable_type
class Foo: pass


def test_get_variable_type():
    """
    Tests for ``otter.utils.get_variable_type``.
    """
    assert get_variable_type(Foo()) == "test.test_utils.Foo"
    assert get_variable_type(pd.DataFrame()) == "pandas.core.frame.DataFrame"



"""Tests for ``otter.run``"""

import copy
import json
import nbformat
import nbconvert
import os
import pytest
import re
import shutil

from contextlib import contextmanager, nullcontext
from textwrap import dedent
from unittest import mock

from otter.generate.token import APIClient
from otter.run.run_autograder import main as run_autograder
from otter.run.run_autograder.utils import OtterRuntimeError
from otter.utils import NBFORMAT_VERSION

from ..utils import delete_paths, TestFileManager


FILE_MANAGER = TestFileManager(__file__)


@pytest.fixture(autouse=True)
def cleanup_output(cleanup_enabled):
    with FILE_MANAGER.open("autograder/source/otter_config.json") as f:
        cpy = f.read()
    with FILE_MANAGER.open("rmd-autograder/source/otter_config.json") as f:
        crmd = f.read()
    yield
    if cleanup_enabled:
        delete_paths([
            FILE_MANAGER.get_path("autograder/results/results.json"),
            FILE_MANAGER.get_path("autograder/results/results.pkl"),
            FILE_MANAGER.get_path("autograder/__init__.py"),
            FILE_MANAGER.get_path("autograder/submission/test"),
            FILE_MANAGER.get_path("autograder/submission/tests"),
            FILE_MANAGER.get_path("autograder/submission/fails2and6H.pdf"),
            FILE_MANAGER.get_path("autograder/submission/__init__.py"),
            FILE_MANAGER.get_path("autograder/submission/.OTTER_LOG"),
            FILE_MANAGER.get_path("rmd-autograder/results/results.json"),
            FILE_MANAGER.get_path("rmd-autograder/results/results.pkl"),
            FILE_MANAGER.get_path("rmd-autograder/__init__.py"),
            FILE_MANAGER.get_path("rmd-autograder/submission/test"),
            FILE_MANAGER.get_path("rmd-autograder/submission/tests"),
            FILE_MANAGER.get_path("rmd-autograder/submission/__init__.py"),
            FILE_MANAGER.get_path("rmd-autograder/submission/.OTTER_LOG"),
        ])
        with FILE_MANAGER.open("autograder/source/otter_config.json", "w") as f:
            f.write(cpy)
        with FILE_MANAGER.open("rmd-autograder/source/otter_config.json", "w") as f:
            f.write(crmd)


@pytest.fixture(autouse=True)
def correct_cwd_on_exit():
    cwd = os.getcwd()
    yield
    os.chdir(cwd)


@pytest.fixture
def expected_results():
    return {
        "tests": [
            {
                "name": "Public Tests",
                "visibility": "visible",
                "output": "q1 results: All test cases passed!\n\nq2 results:\n    q2 - 1 result:\n        ❌ Test case failed\n        Trying:\n            negate(True)\n        Expecting:\n            False\n        **********************************************************************\n        Line 2, in q2 0\n        Failed example:\n            negate(True)\n        Expected:\n            False\n        Got:\n            True\n\n    q2 - 2 result:\n        ❌ Test case failed\n        Trying:\n            negate(False)\n        Expecting:\n            True\n        **********************************************************************\n        Line 2, in q2 1\n        Failed example:\n            negate(False)\n        Expected:\n            True\n        Got:\n            False\n\nq3 results: All test cases passed!\n\nq4 results: All test cases passed!\n\nq6 results: All test cases passed!\n\nq7 results: All test cases passed!",
                "status": "failed",
            },
            {
                "name": "q1",
                "score": 0.0,
                "max_score": 0.0,
                "visibility": "hidden",
                "output": "q1 results: All test cases passed!"
            },
            {
                "name": "q2",
                "score": 0,
                "max_score": 2.0,
                "visibility": "hidden",
                "output": "q2 results:\n    q2 - 1 result:\n        ❌ Test case failed\n        Trying:\n            negate(True)\n        Expecting:\n            False\n        **********************************************************************\n        Line 2, in q2 0\n        Failed example:\n            negate(True)\n        Expected:\n            False\n        Got:\n            True\n\n    q2 - 2 result:\n        ❌ Test case failed\n        Trying:\n            negate(False)\n        Expecting:\n            True\n        **********************************************************************\n        Line 2, in q2 1\n        Failed example:\n            negate(False)\n        Expected:\n            True\n        Got:\n            False\n\n    q2 - 3 result:\n        ❌ Test case failed\n        Trying:\n            negate(\"\")\n        Expecting:\n            True\n        **********************************************************************\n        Line 2, in q2 2\n        Failed example:\n            negate(\"\")\n        Expected:\n            True\n        Got:\n            ''\n\n    q2 - 4 result:\n        ❌ Test case failed\n        Trying:\n            negate(1)\n        Expecting:\n            False\n        **********************************************************************\n        Line 2, in q2 3\n        Failed example:\n            negate(1)\n        Expected:\n            False\n        Got:\n            1"
            },
            {
                "name": "q3",
                "score": 2.0,
                "max_score": 2.0,
                "visibility": "hidden",
                "output": "q3 results: All test cases passed!"
            },
            {
                "name": "q4",
                "score": 1.0,
                "max_score": 1.0,
                "visibility": "hidden",
                "output": "q4 results: All test cases passed!"
            },
            {
                "name": "q6",
                "score": 2.5,
                "max_score": 5.0,
                "visibility": "hidden",
                "output": "q6 results:\n    q6 - 1 result:\n        ✅ Test case passed\n\n    q6 - 2 result:\n        ❌ Test case failed\n        Trying:\n            fib = fiberator()\n        Expecting nothing\n        ok\n        Trying:\n            for _ in range(10):\n                print(next(fib))\n        Expecting:\n            0\n            1\n            1\n            2\n            3\n            5\n            8\n            13\n            21\n            34\n        **********************************************************************\n        Line 3, in q6 1\n        Failed example:\n            for _ in range(10):\n                print(next(fib))\n        Expected:\n            0\n            1\n            1\n            2\n            3\n            5\n            8\n            13\n            21\n            34\n        Got:\n            0\n            1\n            1\n            1\n            2\n            3\n            5\n            8\n            13\n            21"
            },
            {
                "name": "q7",
                "score": 1.0,
                "max_score": 1.0,
                "visibility": "hidden",
                "output": "q7 results: All test cases passed!"
            }
        ],
        "output": "Students are allowed 1 submissions every 1 days. You have 0 submissions in that period.",
    }


@pytest.fixture(scope="module")
def expected_rmd_results():
    return {
        "tests": [
            {
                "name": "Public Tests",
                "output": "q1 results: All test cases passed!",
                "visibility": "visible",
                "status": "passed",
            },
            {
                "max_score": 5,
                "name": "q1",
                "output": "q1 results: All test cases passed!\nq1d message: congrats",
                "score": 5,
                "visibility": "hidden",
            },
        ],
    }


@contextmanager
def alternate_config(config_path, new_config):
    with open(config_path) as f:
        contents = f.read()
    with open(config_path, "w") as f:
        json.dump(new_config, f)
    try: 
        yield
    finally:
        with open(config_path, "w") as f:
            f.write(contents)


def get_expected_error_results(error):
    return {
        "score": 0,
        "stdout_visibility": "hidden",
        "tests": [
            {
                "name": "Autograder Error",
                "output": f"Otter encountered an error when grading this submission:\n\n{error}",
            },
        ],
    }


@contextmanager
def alternate_submission(subm_path, new_nb):
    with open(subm_path) as f:
        contents = f.read()
    nbformat.write(new_nb, subm_path)
    try: 
        yield
    finally:
        with open(subm_path, "w") as f:
            f.write(contents)


@contextmanager
def alternate_tests(test_files):
    src, dst = FILE_MANAGER.get_path("autograder/source/tests"), FILE_MANAGER.get_path("autograder/source/tests_orig")
    os.rename(src, dst)
    os.makedirs(src, exist_ok=True)
    for fn, contents in test_files.items():
        with open(os.path.join(src, fn), "w+") as f:
            f.write(contents)
    try: 
        yield
    finally:
        shutil.rmtree(src)
        os.rename(dst, src)


@pytest.fixture(autouse=True)
def mock_export_notebook(cleanup_enabled):
    empty_pdfs = []
    def make_empty_pdf(*args, **kwargs):
        p = kwargs["dest"]
        open(p, "wb+").close()
        empty_pdfs.append((os.getcwd(), p))

    with mock.patch("otter.run.run_autograder.runners.python_runner.export_notebook") as mocked_export_notebook:
        mocked_export_notebook.side_effect = make_empty_pdf
        yield

    if not cleanup_enabled: return

    for wd, p in empty_pdfs:
        full_path = os.path.join(wd, p)
        if os.path.exists(full_path): os.remove(full_path)


@pytest.fixture
def get_config_path():
    def do_get_config_path(rmd=False):
        dirname = "autograder" if not rmd else "rmd-autograder"
        return FILE_MANAGER.get_path(f"{dirname}/source/otter_config.json")
    return do_get_config_path


@pytest.fixture
def load_config(get_config_path):
    def load_config_file(rmd=False):
        with open(get_config_path(rmd=rmd)) as f:
            return json.load(f)
    return load_config_file


def test_notebook(load_config, expected_results):
    config = load_config()
    run_autograder(config['autograder_dir'])

    with FILE_MANAGER.open("autograder/results/results.json") as f:
        actual_results = json.load(f)

    assert actual_results == expected_results, \
        f"Actual results did not matched expected:\n{actual_results}"


@mock.patch("otter.run.run_autograder.runners.python_runner.export_notebook")
def test_pdf_generation_failure(mocked_export, get_config_path, load_config, expected_results):
    config = load_config()
    config["warn_missing_pdf"] = True
    config["token"] = "abc123"

    expected_results["tests"].insert(1, {
        "name": "PDF Generation Failed",
        "visibility": "visible",
        "output": "nu-uh",
        "status": "failed",
    })

    with alternate_config(get_config_path(), config):
        mocked_export.side_effect = ValueError("nu-uh")
        run_autograder(config['autograder_dir'])

    with FILE_MANAGER.open("autograder/results/results.json") as f:
        actual_results = json.load(f)

    assert actual_results == expected_results, \
        f"Actual results did not matched expected:\n{actual_results}"


@mock.patch.object(APIClient, "upload_pdf_submission")
@mock.patch("otter.run.run_autograder.runners.python_runner.export_notebook")
def test_use_submission_pdf(
    mocked_export,
    mocked_upload_pdf_submission,
    get_config_path,
    load_config,
    expected_results,
):
    config = load_config()
    config["use_submission_pdf"] = True
    config["token"] = "abc123"

    FILE_MANAGER.open("autograder/submission/fails2and6H.pdf", "wb+").close()

    with alternate_config(get_config_path(), config):
        run_autograder(config['autograder_dir'])

    with FILE_MANAGER.open("autograder/results/results.json") as f:
        actual_results = json.load(f)

    mocked_export.assert_not_called()
    mocked_upload_pdf_submission.assert_called_with(
        config["course_id"],
        config["assignment_id"],
        "student@univ.edu", # from submission_metadata.json in autograder dir
        "fails2and6H.pdf",
    )
    assert actual_results == expected_results, \
        f"Actual results did not matched expected:\n{actual_results}"


def test_force_public_test_summary(get_config_path, load_config):
    def perform_test(show_hidden, force_public_test_summary, expect_summary):
        config = load_config()
        config["show_hidden"] = show_hidden
        config["force_public_test_summary"] = force_public_test_summary
        with alternate_config(get_config_path(), config):
            run_autograder(config['autograder_dir'])

        with FILE_MANAGER.open("autograder/results/results.json") as f:
            actual_results = json.load(f)

        message = f"show_hidden={show_hidden}, force_public_test_summary={force_public_test_summary}, expect_summary={expect_summary}"
        if expect_summary:
            assert actual_results["tests"][0]["name"] == "Public Tests", message
        else:
            assert actual_results["tests"][0]["name"] != "Public Tests", message

    perform_test(False, False, True)
    perform_test(False, True, True)
    perform_test(True, False, False)
    perform_test(True, True, True)


def test_script(load_config, expected_results, get_config_path):
    config = load_config()
    nb_path = FILE_MANAGER.get_path("autograder/submission/fails2and6H.ipynb")
    nb = nbformat.read(nb_path, as_version=NBFORMAT_VERSION)

    # Remove the token so that we don't try to generate a PDF of a script.
    config.pop("token")

    try:
        py, _ = nbconvert.export(nbconvert.PythonExporter, nb)

        # remove magic commands
        py = "\n".join(l for l in py.split("\n") if not l.startswith("get_ipython"))

        os.remove(nb_path)
        with FILE_MANAGER.open("autograder/submission/fails2and6H.py", "w+") as f:
            f.write(py)

        with alternate_config(get_config_path(), config):
            run_autograder(config['autograder_dir'])

        with FILE_MANAGER.open("autograder/results/results.json") as f:
            actual_results = json.load(f)

        assert actual_results == expected_results, \
            f"Actual results did not matched expected:\n{actual_results}"

    finally:
        delete_paths([FILE_MANAGER.get_path("autograder/submission/fails2and6H.py")])
        with open(nb_path, "w+") as f:
            nbformat.write(nb, f)


def test_assignment_name(load_config, expected_results):
    name = "hw01"
    config = load_config()
    nb_path = FILE_MANAGER.get_path("autograder/submission/fails2and6H.ipynb")
    orig_nb = nbformat.read(nb_path, as_version=NBFORMAT_VERSION)
    nb = copy.deepcopy(orig_nb)

    def perform_test(nb, expected_results, error=None, **kwargs):
        nbformat.write(nb, nb_path)

        cm = pytest.raises(OtterRuntimeError, match=re.escape(error)) if error is not None \
            else nullcontext()
        with cm:
            run_autograder(config["autograder_dir"], assignment_name = name, **kwargs)

        with FILE_MANAGER.open("autograder/results/results.json") as f:
            actual_results = json.load(f)

        assert actual_results == expected_results, \
            f"Actual results did not matched expected:\n{actual_results}"

    error_message_template = "Received submission for assignment '{got}' (this is assignment " \
            "'{want}')"

    try:
        # test with correct name
        nb["metadata"]["otter"] = {"assignment_name": name}
        perform_test(nb, expected_results)

        # test with wrong name
        bad_name = "lab01"
        error_message = error_message_template.format(got=bad_name, want=name)
        nb["metadata"]["otter"]["assignment_name"] = bad_name
        perform_test(nb, get_expected_error_results(error_message), error=error_message)

        # test with no name in nb
        error_message = error_message_template.format(got=None, want=name)
        nb["metadata"]["otter"].pop("assignment_name")
        perform_test(nb, get_expected_error_results(error_message), error=error_message)

    finally:
        delete_paths([nb_path])
        with open(nb_path, "w+") as f:
            nbformat.write(orig_nb, f)


def test_rmd(load_config, expected_rmd_results):
    name = "hw01"
    config = load_config(True)
    rmd_path = FILE_MANAGER.get_path("rmd-autograder/submission/hw01.Rmd")
    with open(rmd_path) as f:
        orig_rmd = f.read()

    sub_name = lambda n: re.sub(r"assignment_name: \"\w+\"", f"assignment_name: \"{n}\"", orig_rmd)

    def perform_test(rmd, expected_results, error=None, **kwargs):
        with open(rmd_path, "w") as f:
            f.write(rmd)

        cm = pytest.raises(OtterRuntimeError, match=re.escape(error)) if error is not None \
            else nullcontext()
        with cm:
            run_autograder(config["autograder_dir"], assignment_name = name, **kwargs)

        with FILE_MANAGER.open("rmd-autograder/results/results.json") as f:
            actual_results = json.load(f)

        assert actual_results == expected_results, \
            f"Actual results did not matched expected:\n{actual_results}"

    error_message_template = "Received submission for assignment '{got}' (this is assignment " \
            "'{want}')"

    try:
        # test with correct name
        perform_test(orig_rmd, expected_rmd_results)

        # test with wrong name
        bad_name = "lab01"
        error_message = error_message_template.format(got=bad_name, want=name)
        perform_test(
            sub_name(bad_name), get_expected_error_results(error_message), error=error_message)

        # test with no name in Rmd
        error_message = error_message_template.format(got=None, want=name)
        perform_test(
            "\n".join([l for l in orig_rmd.split("\n") if not l.startswith("assignment_name: ")]),
            get_expected_error_results(error_message),
            error=error_message,
        )

    finally:
        delete_paths([rmd_path])
        with open(rmd_path, "w+") as f:
            f.write(orig_rmd)


@mock.patch.object(APIClient, "upload_pdf_submission")
def test_token_sanitization(mocked_upload, get_config_path, load_config, expected_results):
    """
    Tests that the PDF upload token can't be accessed by the submission.
    """
    config = load_config()
    config["token"] = "abc123"
    config.pop("plugins")

    nb = nbformat.v4.new_notebook(cells=[nbformat.v4.new_code_cell(dedent("""\
        import json
        with open("../source/otter_config.json") as f:
            config = json.load(f)
        token = config.get("token")
    """))])

    tests = {"q1.py": dedent("""\
        OK_FORMAT = False

        from otter.test_files import test_case

        name = "q1"

        @test_case()
        def q1_1(env):
            assert env["token"] is None
    """)}

    expected_results = {
        "tests": [
            {
                "name": "Public Tests",
                "visibility": "visible",
                "output": "q1 results: All test cases passed!",
                "status": "passed",
            },
            {
                "name": "q1",
                "score": 1.0,
                "max_score": 1.0,
                "visibility": "hidden",
                "output": "q1 results: All test cases passed!"
            },
        ],
    }

    FILE_MANAGER.open("autograder/submission/fails2and6H.pdf", "wb+").close()

    mocked_upload.return_value.status_code = 200

    with alternate_config(get_config_path(), config), \
            alternate_submission(FILE_MANAGER.get_path("autograder/submission/fails2and6H.ipynb"), nb), \
            alternate_tests(tests):
        run_autograder(config['autograder_dir'])

    with FILE_MANAGER.open("autograder/results/results.json") as f:
        actual_results = json.load(f)

    assert actual_results == expected_results, \
        f"Actual results did not matched expected:\n{actual_results}"

    mocked_upload.assert_called()
